"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_base models *

:details: lara_django_base database models.
         -
         - fixtures can be generated with
         lara-django-dev dumpdata lara_django_base.itemstatus --indent 2 --natural-primary > 2_itemstatus_fix.json

         model_curi are compact URIs (CURIs) for the model, e.g. lara:base/ItemStatus.
         They are used for semantic representation.

:authors: mark doerr <mark.doerr@uni-greifswald.de>
          David Lacko @deiwo
          Jan Velecký
________________________________________________________________________
"""

import hashlib
import json
import uuid
from typing import Any, ClassVar
from urllib.parse import urlparse

from django.conf import settings
from django.contrib.postgres.indexes import GinIndex
from django.contrib.postgres.search import SearchVectorField
from django.core.validators import RegexValidator
from django.db import models
from django.utils import timezone

settings.FIXTURES += [
    "1_currency_fix",
    "2_datatypes_fix",
    "2_itemstatus_fix",
    "2_tags_fix",
    "5_namespace_fix",
    "6_iana_media_type_fix",
    "9_physical_unit_fix",
    "9_physical_state_matter_fix",
    "9_scientific_constants",
]


class URNField(models.CharField):
    """URN field that accepts URLs that adhere the scheme for URNs  only."""

    default_validators : ClassVar[list[RegexValidator]] = [
        RegexValidator(
            regex=r"^urn:[a-z0-9][a-z0-9-]{1,31}:([a-z0-9()+,-.:=@;$_!*\'%/?#]|%[0-9a-f][0-9a-f])*$",
            message="Please enter a valid URN, s. https://en.wikipedia.org/wiki/Uniform_Resource_Name",
            code="invalid_URN",
        )
    ]  # [URLValidator(schemes=['urn'])]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.max_length = 512


class MediaType(models.Model):
    """
    Generic media type (IANA MIME system).

    See https://www.iana.org/assignments/media-types/media-types.xhtml for media files, like
    e.g. svg, JSON, png, mp3, pdf, ...
    Format:
     type / subtype *(; parameter) application/x-animl, application/json
    """

    model_curi = "lara:base/MediaType"

    mediatype_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    name = models.TextField(help_text="name of the IANA application")
    # TODO: check for uniqueness of csv file
    media_type = models.TextField(
        unique=True,
        help_text="""name of IANA/MIME media type in the format:
                    type '/' [tree '.'] subtype ['+' suffix] *[';' parameter]""",
    )  # unique=True,
    def_filename_extension_regex = models.TextField(
        blank=True,
        unique=True,
        help_text=r"""default filename extension regular expression,
                     e.g. ^.*\ .(jpg|JPG)$ ,  ^.*\ .(gif|GIF|doc|DOC|pdf|PDF)$""",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    description = models.TextField(blank=True, help_text="description of file type")

    class Meta:
        """Model metadata."""

        ordering: ClassVar[list[str]] = ["name"]
        indexes: ClassVar[list[models.Index]] = [
            models.Index(fields=["name", "media_type"], name="name_media_type_idx")
        ]
        db_table_comment = (
            "Generic media Type (IANA type/subtype) for media files, like, "
            "e.g., application/text, application/json, image/svg+xml, image/png, audio/mp3, application/pdf, "
        )

    def __str__(self):
        """Return the media type string."""
        return self.media_type or ""

    def __repr__(self):
        """Return the media type representation."""
        return self.media_type or ""

    def save(self, *args, **kwargs):
        """Generate default values for the IRI."""
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/MediaType/"
                + self.name.replace(" ", "_").lower().strip()
            )
        if self.mediatype_id is None:
            if self.media_type is not None:
                self.mediatype_id = uuid.uuid5(uuid.NAMESPACE_URL, self.media_type)
            else:
                self.mediatype_id = uuid.uuid4()

        super().save(*args, **kwargs)


class DataType(models.Model):
    """
    Generic DataType that can be used to classify ExtraData (s. below).

    e.g. email_home, email_lab, telephone_number_home1, company_x_customer_number,
    sequence data, absorption data ...
    """

    model_curi = "lara:base/DataType"

    datatype_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.TextField(
        unique=True, db_index=True, help_text="name of the data type"
    )
    media_type = models.ForeignKey(
        MediaType,
        related_name="%(app_label)s_%(class)s_file_types_related",
        related_query_name="%(app_label)s_%(class)s_file_types_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="IANA media type of extra data file",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        blank=True, help_text="description of extra data type"
    )

    class Meta:
        """Model metadata."""

        ordering: ClassVar[list[str]] = ["name"]
        db_table_comment = (
            "Generic DataType that can be used to classify Data (s. below) e.g. email_home,"
            "email_lab, telephone_number_home1, company_x_customer_number, sequence data, absorption data ..."
        )

    def __str__(self):
        """Return the data type name."""
        return self.name or ""

    def __repr__(self):
        """Return the data type representation."""
        return self.name or ""

    def save(self, *args, **kwargs):
        """Generate default values for the IRI."""
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/DataType/"
                + self.name.replace(" ", "_").lower().strip()
            )
        if self.datatype_id is None:
            if self.name is not None:
                self.datatype_id = uuid.uuid5(uuid.NAMESPACE_URL, self.name)
            else:
                self.datatype_id = uuid.uuid4()

        super().save(*args, **kwargs)


class Namespace(models.Model):
    """
    Generic namespace model.

    Namespaces can be used to specify the scope or provenience of data, entities, etc. and help to make
    them unique.
    """

    model_curi = "lara:base/Namespace"
    namespace_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    uri = models.URLField(
        unique=True,
        db_index=True,
        help_text="Universal Resource Identifier - A unique namespace for identifying data, https://de.unigreifswald/biochem/akb",
    )
    description = models.TextField(
        blank=True, help_text="description of this Namespace"
    )

    # search_vector = SearchVectorField(null=True)

    class Meta:
        """Model metadata."""

        ordering: ClassVar[list[str]] = ["uri"]
        db_table_comment = (
            "Generic Namespace model: Namespaces can be used to specify the scope or "
            "provenience of data, entities, etc. and helps to make them unique."
        )
        # indexes = (GinIndex(fields=["search_vector"]),)

    def __str__(self):
        """Return the namespace URI."""
        return self.uri or ""

    def __repr__(self):
        """Return the namespace representation."""
        return self.uri or ""

    def save(self, *args, **kwargs):
        """Generate default values for the id."""
        if self.namespace_id is None:
            if self.uri is not None:
                self.namespace_id = uuid.uuid5(uuid.NAMESPACE_URL, self.uri)
            else:
                self.namespace_id = uuid.uuid4()

        super().save(*args, **kwargs)

    def parse_uri(self):
        """
        Parse a URI into parts.

        To retrieve the parts of the URI, use the following attributes (example: https://de.unigreifswald/biochem/akb"):
                result = parse_URI()
                    result.scheme : "https"
                    result.netloc : "de.unigreifswald"
                    result.path : "/biochem/akb"
        """
        return urlparse(self.uri)


class Tag(models.Model):
    """
    Generic tags and keywords model.

    Tags and keywords can be used to group related information or make objects findable.
    """

    model_curi = "lara:base/Tag"
    tag_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    tag = models.TextField(
        unique=True,
        db_index=True,
        help_text="a tag, can be used for tagging items, like 'reviewed', 'interesting',"
        " but also for keywords, like 'python', 'evolution'",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    # search_vector = SearchVectorField(null=True)

    class Meta:
        """Model metadata."""

        ordering: ClassVar[list[str]] = ["tag"]
        db_table_comment = (
            "Generic tags and keywords model: Tags and keywords can be used"
            "to group related information or make objects findable."
        )
        # indexes = (GinIndex(fields=["search_vector"]),)

    def __str__(self):
        """Return the tag string."""
        return self.tag or ""

    def save(self, *args, **kwargs):
        """Generate default values for the IRI."""
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/Tag/"
                + self.tag.replace(" ", "_").lower().strip()
            )

        super().save(*args, **kwargs)

class ItemRole(models.Model):
    """
    Generic item role.

    Can be used for indicating the role of LARA items, like containers or processes: sample, reference,
    standard, control, ...
    """

    model_curi = "lara:base/ItemRole"
    itemrole_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    role = models.TextField(
        db_index=True,
        help_text="item role, like sample, reference, standard, control, ..",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        blank=True, help_text="A short description of the item role"
    )

    class Meta:
        """Model metadata."""

        verbose_name_plural = "ItemRoles"
        db_table_comment = (
            "Generic item role/state, can be used for indicating the role of LARA items,"
            "like a containers or processes: sample, reference, standard, control, ..."
        )

    def __str__(self):
        """Return the item role string."""
        return self.role or ""

class ItemStatus(models.Model):
    """
    Generic item status/state for LARA items.

    Indicates status for containers or processes (planned, started, evaluating) or order states, like
    to_order, pending, ordered, received ...
    Hint: please do not use tags for this - separating tag and status makes processing clearer.
    """

    model_curi = "lara:base/ItemStatus"
    itemstatus_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    status = models.TextField(
        db_index=True,
        help_text="item status, like planned, started, evaluating, .. finished",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    color = models.TextField(
        # default=ColorBlind16.full_color_scheme[randint(0, 16)].get_web(),
        blank=True,
        help_text="colour scheme indicating the status of the item",
    )
    value = models.FloatField(
        default=0.0,
        blank=True,
        help_text="relative value between 0 and 1, indicating the status of the item",
    )
    description = models.TextField(
        blank=True, help_text="A short description of the item status"
    )
    tags = models.ManyToManyField(
        Tag,
        related_name="%(app_label)s_%(class)s_tags_related",
        related_query_name="%(app_label)s_%(class)s_tags_related_query",
        blank=True,
        help_text="tags for item status",
    )

    class Meta:
        """Model metadata."""

        verbose_name_plural = "ItemStatus"
        db_table_comment = (
            "Generic item status/state, can be used for indicating the status of LARA items,"
            "like a containers or processes: planned, started, evaluating or order states, like to_order,"
            "pending, ordered, received ..."
        )

    def __str__(self):
        """Return the item status string."""
        return self.status or ""

    def save(self, *args, **kwargs):
        """Generate default values for the IRI."""
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/ItemStatus/"
                + self.status.replace(" ", "_").lower().strip()
            )

        super().save(*args, **kwargs)


# TODO: add a subset for status

# class ItemRole(models.Model):
#     """generic item role, can be used for indicating the role of LARA items,
#     like a containers or processes: sample, reference, standard, control, ...
#     """

#     model_curi = "lara:base/ItemRole"
#     itemrole_id = models.UUIDField(
#         primary_key=True, default=uuid.uuid4, editable=False
#     )
#     role = models.TextField(
#         db_index=True,
#         help_text="item role, like sample, reference, standard, control, ..",
#     )
#     iri = models.URLField(
#         blank=True,
#         null=True,
#         unique=True,
#         max_length=512,
#         help_text="International Resource Identifier - IRI: is used for semantic representation ",
#     )
#     description = models.TextField(
#         blank=True, help_text="A short description of the item role"
#     )

#     class Meta:
#         verbose_name_plural = "ItemRoles"
#         db_table_comment = (
#             "Generic item role/state, can be used for indicating the role of LARA items,"
#             "like a containers or processes: sample, reference, standard, control, ..."
#         )

#     def __str__(self):
#         return self.role or ""

#     def save(self, *args, **kwargs):
#         """
#         Here we generate some default values for the IRI
#         """
#         if not self.iri:
#             self.iri = (
#                 settings.LARA_PREFIXES["lara"]
#                 + "base/ItemRole/"
#                 + self.role.replace(" ", "_").lower().strip()
#             )

#         super().save(*args, **kwargs)


class BlockchainHashesAbstr(models.Model):
    """Store the latest blockchain hash of a data block to build the chain."""

    blockchainhashes_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )

    block_id = models.TextField(
        blank=True,
        help_text="Identifier of the data block for which the blockchain hash was currently calculated,"
        "e.g., data of a measurement series.",
    )

    hash_sha256 = models.CharField(
        max_length=256,
        unique=True,
        blank=True,
        null=True,
        default=None,
        help_text="SHA256 hash of all data (JSON and XML)",
    )
    hash_sha512 = models.CharField(
        max_length=512,
        unique=True,
        blank=True,
        null=True,
        default=None,
        help_text="SHA512 hash of all data (JSON and XML)",
    )
    timestamp = models.DateTimeField(
        default=timezone.now,
        help_text="date and time when the blockchain hash was last modified",
    )

    class Meta:
        """Model metadata."""

        abstract = True
        verbose_name_plural = "BlockchainHashes"

    def save(self, *args, **kwargs):
        """Generate default values for the IRI."""
        self.timestamp = timezone.now()

        super().save(*args, **kwargs)


class ExtraDataAbstr(models.Model):
    """
    Extend data with extra information.

    e.g. more telephone numbers, customer numbers, ...
    """

    extradata_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    data_type = models.ForeignKey(
        DataType,
        related_name="%(app_label)s_%(class)s_name_related",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    namespace = models.ForeignKey(
        Namespace,
        related_name="%(app_label)s_%(class)s_namespaces_related",
        related_query_name="%(app_label)s_%(class)s_namespaces_related_query",
        on_delete=models.CASCADE,
        help_text="namespace of data",
    )
    name_display = models.TextField(
        blank=True,
        help_text="Human readable display name or title of extra data",
    )
    name = models.TextField(blank=True, help_text="name of extra data")
    name_full = models.TextField(
        unique=True, blank=True, help_text="full name of extra data"
    )

    version = models.TextField(blank=True, help_text="version of extra data")

    hash_sha256 = models.CharField(
        max_length=256,
        unique=True,
        blank=True,
        default=None,
        help_text="SHA256 hash of all data (JSON and XML)",
    )
    data_json = models.JSONField(blank=True, null=True, help_text="generic JSON field")
    media_type = models.ForeignKey(
        MediaType,
        related_name="%(app_label)s_%(class)s_file_types_related",
        related_query_name="%(app_label)s_%(class)s_file_types_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="IANA media type of extra data file",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    url = models.URLField(
        blank=True,
        help_text="Universal Resource Locator - this can be used to link the data to external locations",
    )
    datetime_created = models.DateTimeField(
        blank=True, null=True, help_text="date and time when data was created"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True, null=True, help_text="date and time when data was last modified"
    )
    description = models.TextField(blank=True, help_text="description of extra data")
    search_vector = SearchVectorField(null=True)

    class Meta:
        """Model metadata."""

        abstract = True
        verbose_name_plural = "ExtraData"
        constraints: ClassVar[list[models.UniqueConstraint]] = [
            models.UniqueConstraint(
                fields=["name", "namespace", "version"],
                name="%(app_label)s_%(class)s_name_namespace_version_unique",
            ),
        ]
        indexes: ClassVar[tuple[models.Index, ...]] = (GinIndex(fields=["search_vector"]),)

    def __str__(self):
        """Return the extra data name."""
        return f"{self.name_full}" or ""

    def __repr__(self):
        """Return the extra data representation."""
        return f"{self.name_full}" or ""


class ExtraData(ExtraDataAbstr):
    """
    Extend data with extra information.

    e.g., more telephone numbers, customer numbers, ...
    """

    model_curi = "lara:base/ExtraData"
    # extradata_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    file = models.FileField(
        upload_to="base", blank=True, null=True, help_text="rel. path/filename"
    )
    image = models.ImageField(
        upload_to="base/images/",
        blank=True,
        null=True,
        help_text="location room map rel. path/filename to image",
    )

    class Meta:
        """Model metadata."""

        indexes = (
            models.Index(fields=["name", "namespace"], name="name_namespace_idx"),
        )
        db_table_comment = (
            "This class can be used to extend data, by extra information,"
            "e.g. more telephone numbers, customer numbers, ..."
        )
        verbose_name_plural = "ExtraData"

    # def save(self, force_insert=None, using=None): #force_insert=force_insert, using=using
    def save(self, *args, **kwargs):
        """Generate default values for name and name_full."""
        if self.version is None or self.version == "":
            self.version = "0.0.1"

        if self.name is None or self.name == "":
            if self.name_display is None or self.name_display == "":
                self.name = "_".join(
                    (
                        timezone.now().strftime("%Y%m%d_%H%M%S"),
                        "data",
                        self.hash_sha256[:8],
                    )
                )
            else:
                self.name = self.name_display.replace(" ", "_").lower().strip()

        else:
            self.name = self.name.replace(" ", "_").lower().strip()
        if self.name_full is None or self.name_full == "":
            if self.namespace is not None and self.version is not None:
                self.name_full = f"{self.namespace.uri}/" + "_".join(
                    (self.name.replace(" ", "_"), self.version)
                )
            else:
                self.name_full = "_".join((self.name.replace(" ", "_"), self.version))

        if not self.iri:
            netloc = self.namespace.parse_uri().netloc
            path = self.namespace.parse_uri().path
            self.iri = f"{settings.LARA_PREFIXES['lara']}base/ExtraData/{netloc}{path}/{self.name.replace(' ', '_').lower().strip()}"  # noqa: E501

        if self.hash_sha256 is None:
            if self.data_json is not None:
                to_hash = json.dumps(self.data_json)
                self.hash_sha256 = hashlib.sha256(to_hash.encode("utf-8")).hexdigest()
            else:
                self.hash_sha256 = hashlib.sha256(
                    str(self.iri).encode("utf-8")
                ).hexdigest()

        self.datetime_created = (
            timezone.now() if self.datetime_created is None else self.datetime_created
        )

        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)


class ExternalResource(models.Model):
    """
    Generic external resource class.

    External resources can be used to link to external resources, like e.g. literature, pdf, a website,
    a file, a database, ...
    """

    model_curi = "lara:base/ExternalResource"
    externalresource_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    namespace = models.ForeignKey(
        Namespace,
        related_name="%(app_label)s_%(class)s_namespaces_related",
        related_query_name="%(app_label)s_%(class)s_namespaces_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="namespace of external resource",
    )
    name_display = models.TextField(
        blank=True,
        help_text="Human readable display name or title of external resource",
    )
    name = models.TextField(
        unique=True, blank=True, help_text="name of the external resource"
    )
    resource_type = models.ForeignKey(
        DataType,
        related_name="%(app_label)s_%(class)s_resource_type_related",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="type of external resource, like pdf, website, handle, DOI, ...",
    )
    url = models.URLField(
        blank=True,
        help_text="Universal Resource Locator - URL - this can be used to link the data to external locations",
    )
    iri = models.URLField(
        blank=True,
        unique=True,
        help_text="Universal Resource Name - URI - this can be used to uniquely identify the external resource",
    )
    prefix = models.TextField(
        blank=True,
        help_text="prefix for compact URIs (CURI), e.g. 'skos' for 'http://www.w3.org/2004/02/skos/core#'",
    )
    tags = models.ManyToManyField(
        Tag,
        related_name="%(app_label)s_%(class)s_tags_related",
        related_query_name="%(app_label)s_%(class)s_tags_related_query",
        blank=True,
        help_text="tags for external resource",
    )
    description = models.TextField(
        blank=True, help_text="description of the external resource"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True, null=True, help_text="date and time when data was last modified"
    )
    search_vector = SearchVectorField(null=True)

    class Meta:
        """Model metadata."""

        indexes = (
            models.Index(fields=["name", "namespace"], name="ext_res_name_namespace_idx"),
            GinIndex(fields=["search_vector"]),
        )
        db_table_comment = (
            "Generic external resource class: External resources can be used "
            "to link to external resources, like e.g. literature, pdf, a website, a file, a database, ..."
        )

    def __str__(self):
        """Return the external resource name."""
        return self.name or ""

    def __repr__(self):
        """Return the external resource representation."""
        return self.name or ""

    def save(self, *args, **kwargs):
        """Generate default values for the name."""
        if self.name_display is not None or "":
            if self.name is None or self.name == "":
                self.name = self.name_display.replace(" ", "_").lower().strip()
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/ExternalResource/"
                + self.name.replace(" ", "_").lower().strip()
            )
        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)


class PhysicalUnit(models.Model):
    """Physical units, like meter, kilogram, ..., preferably SI units."""

    model_curi = "lara:base/PhysicalUnit"
    physicalunit_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    name = models.TextField(help_text="physical unit, like meter, kilogram, ...")
    symbol = models.TextField(
        blank=True, help_text="symbol of the physical unit, like m, kg, ..."
    )
    dimension = models.TextField(
        blank=True, help_text="physical dimension, like length, mass, ..."
    )
    transformation_si = models.JSONField(
        blank=True,
        null=True,
        help_text="transformation formula to SI unit, in machine readable form (e.g. python code), like: x * 1000.0",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        blank=True, help_text="A short description of the physical unit"
    )

    class Meta:
        """Model metadata."""

        db_table_comment = (
            "Physical units, like, meter, kilogram, ..., preferably SI units"
        )

    def __str__(self):
        """Return the physical unit name."""
        return self.name or ""

    def __repr__(self):
        """Return the physical unit representation."""
        return self.name or ""

    def save(self, *args, **kwargs):
        """Generate default values for the IRI."""
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/PhysicalUnit/"
                + self.name.replace(" ", "_").lower().strip()
            )

        super().save(*args, **kwargs)


class PhysicalStateMatter(models.Model):
    """Physical state of matter, like solid, liquid, gas, plasma, ... ."""

    model_curi = "lara:base/PhysicalStateMatter"
    physicalstatematter_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    name = models.TextField(
        help_text="physical state of matter, like solid, liquid, gas, plasma, ..."
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        blank=True,
        help_text="A short description of the physical state of matter",
    )

    class Meta:
        """Model metadata."""

        db_table_comment = (
            "Physical state of matter, like solid, liquid, gas, plasma, ..."
        )
        verbose_name_plural = "PhysicalStatesMatter"

    def __str__(self):
        """Return the physical state name."""
        return self.name or ""

    def __repr__(self):
        """Return the physical state representation."""
        return self.name or ""


class ScientificConstant(models.Model):
    """Mathematical and physical constants."""

    model_curi = "lara:base/ScientificConstant"
    scientificconstant_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    name = models.TextField(blank=True, help_text="generic text field")
    symbol = models.TextField(blank=True, help_text="symbol of the constant")
    value = models.FloatField(blank=True, null=True, help_text="Float value of the ")
    uncertainty = models.FloatField(
        blank=True, null=True, help_text="uncertainty of the value +/-"
    )
    unit = models.ForeignKey(
        PhysicalUnit,
        related_name="%(app_label)s_%(class)s_units_related",
        related_query_name="%(app_label)s_%(class)s_units_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="physical unit",
    )
    data_json = models.JSONField(blank=True, null=True, help_text="complex content")
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    # -> external resources
    url_source = models.URLField(
        blank=True,
        help_text="Universal Resource Locator - URL - source, e.g. in literature",
    )
    url_definition = models.URLField(
        blank=True, help_text="Universal Resource Locator - URL - definition"
    )
    description = models.TextField(blank=True, help_text="description of the constant")
    datetime_last_modified = models.DateTimeField(
        blank=True, null=True, help_text="date and time when data was last modified"
    )
    search_vector = SearchVectorField(null=True)

    class Meta:
        """Model metadata."""

        db_table_comment = "Mathematical and Physical constants"
        indexes = (GinIndex(fields=["search_vector"]),)

    def __str__(self):
        """Return the constant name."""
        return f"{self.name}" or ""

    def __repr__(self):
        """Return the constant representation."""
        return f"{self.name} ({self.description})" or ""

    def save(self, *args, **kwargs):
        """Generate default values for the IRI."""
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/ScientificConstant/"
                + self.name.replace(" ", "_").lower().strip()
            )
        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)


class Currency(models.Model):
    """
    Generic currency class.

    Hint: see also Money class from django-shop.
    """

    model_curi = "lara:base/Currency"
    currency_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.TextField(unique=True, help_text="currency name")
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    alpha3code = models.TextField(
        blank=True,
        unique=True,
        help_text="acronym, e.g. EUR, USD, ..",
    )
    numeric_code = models.IntegerField(
        blank=True,
        null=True,
        help_text="International 3 number code, like 978 (for EUR) ",
    )
    symbol = models.TextField(blank=True, help_text="currency symbol, e.g. €, $, ...")
    exchange_rate_eur = models.DecimalField(
        max_digits=20,
        decimal_places=9,
        blank=True,
        null=True,
        help_text="currency exchange rate to Euro ",
    )
    datetime_last_modified = models.DateTimeField(
        default=timezone.now, help_text="date and time when data was last modified"
    )
    description = models.TextField(blank=True, help_text="description of the currency")

    class Meta:
        """Model metadata."""

        verbose_name_plural = "Currencies"
        db_table_comment = "Generic currency class"

    def __str__(self):
        """Return the currency name."""
        return self.name or ""

    def save(self, *args, **kwargs):
        """Generate default values for the IRI."""
        if not self.iri:
            self.iri = (
                settings.LARA_PREFIXES["lara"]
                + "base/Currency/"
                + self.name.replace(" ", "_").lower().strip()
            )

        super().save(*args, **kwargs)

    def to_euro(self, amount: float = 1.0):
        """Convert the given amount to EUR using the stored exchange rate."""
        return amount * self.exchange_rate_eur



class LARAExtension(models.Model):
    """
    Store configuration for LARA web UI extensions.

    Extensions can add new apps, features, or modules as an IFrame or a new page. An (external) URL can be
    provided to point to an external resource or microservice, or a Django app can be added. Extensions can
    be added to the main menu or to a sub-menu.
    """

    model_curi = "lara:base/LARAExtension"
    laraextension_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    name = models.TextField(unique=True, help_text="name of the extension")
    url = models.URLField(blank=True, help_text="External URL of the extension")
    icon = models.TextField(blank=True, help_text="SVG icon of the extension")
    parent = models.ForeignKey(
        "self",
        related_name="%(app_label)s_%(class)s_parents_related",
        related_query_name="%(app_label)s_%(class)s_parents_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="parent extension for building a route / menu hierarchy",
    )
    new_tab = models.BooleanField(default=False, help_text="open in new tab")
    target_element = models.TextField(
        blank=True, help_text="target HTML element of the extension"
    )
    components = models.JSONField(
        blank=True, null=True, help_text="components / templates of the extension"
    )

    description = models.TextField(blank=True, help_text="description of the extension")

    class Meta:
        """Model metadata."""

        db_table_comment = "LARA UI extensions"

    def __str__(self):
        """Return the extension name."""
        return self.name or ""

    def __repr__(self):
        """Return the extension representation."""
        return self.name or ""






from django.conf import settings
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.contrib.postgres.search import SearchVector
from lara_django_base.decorators import postgres_only

from .models.models import (
    ExtraData,
    ExternalResource,
    ScientificConstant,
    City,
    Address,
    Room,
    Location,
)


@receiver(post_save, sender=ExtraData)
@postgres_only
def update_search_vector(sender, instance, **kwargs):
    sender.objects.filter(pk=instance.pk).update(
        search_vector=(
            SearchVector("name", weight="A")
            + SearchVector("name_display", weight="B")
            + SearchVector("description", weight="C")
        )
    )


@receiver(post_save, sender=ExternalResource)
@postgres_only
def update_search_vector(sender, instance, **kwargs):
    sender.objects.filter(pk=instance.pk).update(
        search_vector=(
            SearchVector("name", weight="A") + SearchVector("description", weight="B")
        )
    )


@receiver(post_save, sender=ScientificConstant)
@postgres_only
def update_search_vector(sender, instance, **kwargs):
    sender.objects.filter(pk=instance.pk).update(
        search_vector=(
            SearchVector("name", weight="A") + SearchVector("description", weight="B")
        )
    )


@receiver(post_save, sender=City)
@postgres_only
def update_search_vector(sender, instance, **kwargs):
    sender.objects.filter(pk=instance.pk).update(
        search_vector=(
            SearchVector("name", weight="A")
            + SearchVector("name_local", weight="B")
            + SearchVector("description", weight="C")
        )
    )


@receiver(post_save, sender=Address)
@postgres_only
def update_search_vector(sender, instance, **kwargs):
    sender.objects.filter(pk=instance.pk).update(
        search_vector=(
            SearchVector("name", weight="A") + 
            SearchVector("street", weight="B") +
            SearchVector("description", weight="C")
        )
    )

@receiver(post_save, sender=Room)
@postgres_only
def update_search_vector(sender, instance, **kwargs):
    sender.objects.filter(pk=instance.pk).update(
        search_vector=(
            SearchVector("name", weight="A") + SearchVector("description", weight="B")
        )
    )

@receiver(post_save, sender=Location)
@postgres_only
def update_search_vector(sender, instance, **kwargs):
    sender.objects.filter(pk=instance.pk).update(
        search_vector=(
            SearchVector("name", weight="A") + SearchVector("description", weight="B")
        )
    )
from django.dispatch import receiver
from django.db.models.signals import post_save
from lara_django_base.decorators import postgres_only

from .semantics import insert_semantic_data

logger = logging.getLogger('lara_signals')


# @receiver(post_save, sender=Evaluation)
# def semantics_data_handler(sender, instance, created, **kwargs):
#     """Semantic data post_save signal handler.

#        This function is triggered after an instance of the Evaluation model is saved.
#        It calls the insert_semantics function to handle the semantic data insertion
#        into the triplestore, running asynchronously as celery task.
#     """
#     logger.debug("post_save_handler1: %s, %s ", sender, instance)

#     insert_semantic_data(sender, instance, **kwargs)


# @receiver(post_save, sender=Evaluation)
# @postgres_only
# def update_search_vector(sender, instance, **kwargs):
#     sender.objects.filter(pk=instance.pk).update(
#         search_vector=(
#             SearchVector('name', weight='A') +
#             SearchVector('name_display', weight='B') +
#             SearchVector('description', weight='C') +
#             SearchVector('caption', weight='D')
#         )
#     )


from django import template
from django.utils.safestring import mark_safe
from django.templatetags.static import static

register = template.Library()


@register.simple_tag
def icon(name, class_name="w-4 h-4", title=None, **kwargs):
    """
    Render an SVG icon from the symbols.svg sprite file.
    
    Usage:
        {% icon 'home' %}
        {% icon 'dashboard' class_name='w-6 h-6 text-blue-500' %}
        {% icon 'bell' title='Notifications' %}
    
    Args:
        name (str): The icon name (without prefix)
        class_name (str): CSS classes for styling
        title (str): Optional title for accessibility
        **kwargs: Additional HTML attributes
    
    Returns:
        SafeString: HTML for the SVG icon
    """
    # Build additional attributes
    attrs = []
    for key, value in kwargs.items():
        # Convert Python attribute names to HTML (e.g., data_value -> data-value)
        html_key = key.replace('_', '-')
        attrs.append(f'{html_key}="{value}"')
    
    attrs_str = ' ' + ' '.join(attrs) if attrs else ''
    
    # Add accessibility attributes
    aria_attr = f'aria-label="{title}"' if title else 'aria-hidden="true"'
    
    # Generate the icon HTML
    icon_html = f'''<svg class="{class_name}" {aria_attr}{attrs_str}>
    <use href="{static('lara_django_base/icons/symbols.svg')}#{name}"></use>
</svg>'''
    
    return mark_safe(icon_html)


@register.inclusion_tag('lara_django_base/includes/icon.html')
def icon_include(name, class_name="w-4 h-4", title=None):
    """
    Alternative inclusion tag approach for icons.
    
    Usage:
        {% icon_include 'home' %}
        {% icon_include 'dashboard' class_name='w-6 h-6' %}
    """
    return {
        'name': name,
        'class': class_name,
        'title': title,
    }



# common lara_django decorators

from django.conf import settings

def postgres_only(func):
    def wrapper(*args, **kwargs):
        if settings.DATABASES["default"]["ENGINE"] != "django.db.backends.postgresql":
            return
        return func(*args, **kwargs)
    return wrapper


"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_myproj gRPC handlers*

:details: lara_django_myproj gRPC handlers.
         - 
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - 
________________________________________________________________________
"""

# generated with django-socio-grpc generateprpcinterface lara_django_base  (LARA-version)

# import logging
from django_socio_grpc.services.app_handler_registry import AppHandlerRegistry
from lara_django_base.grpc.services import (
    DataTypeService,
    MediaTypeService,
    NamespaceService,
    NamespaceByURIService,
    TagService,
    ItemStatusService,
    ExtraDataService,
    PhysicalUnitService,
    ExternalResourceService,
    PhysicalStateMatterService,
    ScientificConstantService,
    CurrencyService,
    GeoLocationService,
    CountryService,
    SubdivisionTypeService,
    CountrySubdivisionService,
    CityService,
    AddressService,
    RoomCategoryService,
    RoomService,
    LocationService,
    LARASyncServerService,
)


def grpc_handlers(server):
    app_registry = AppHandlerRegistry("lara_django_base", server)
    # monkey patching to import the grpc module from the API
    app_registry.get_grpc_module = lambda: "lara_django_base_grpc.v1"

    app_registry.register(DataTypeService)

    app_registry.register(MediaTypeService)

    app_registry.register(NamespaceService)

    app_registry.register(NamespaceByURIService)

    app_registry.register(TagService)

    app_registry.register(ItemStatusService)

    app_registry.register(ExtraDataService)

    app_registry.register(PhysicalUnitService)

    app_registry.register(ExternalResourceService)

    app_registry.register(PhysicalStateMatterService)

    app_registry.register(ScientificConstantService)

    app_registry.register(CurrencyService)

    app_registry.register(GeoLocationService)

    app_registry.register(CountryService)

    app_registry.register(SubdivisionTypeService)

    app_registry.register(CountrySubdivisionService)

    app_registry.register(CityService)

    app_registry.register(AddressService)

    app_registry.register(RoomCategoryService)

    app_registry.register(RoomService)

    app_registry.register(LocationService)

    app_registry.register(LARASyncServerService)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Thanks to  Pina Merkert (pmk@ct.de, https://github.com/pinae/Nutria-DB )

from django.core.management.base import BaseCommand
#from django.contrib.auth.models import User
from lara_django_people.models import LaraUser
import os


class Command(BaseCommand):
    def handle(self, *args, **options):
        if LaraUser.objects.count() == 0:
            new_admin = LaraUser.objects.create_superuser(
                username=os.getenv('DJANGO_SUPERUSER_USERNAME', 'admin'),
                email=os.getenv('DJANGO_SUPERUSER_MAIL',
                                'admin@example.com'),
                password=os.getenv('DJANGO_SUPERUSER_PASSWORD', 'yxcv4321'))
            new_admin.save()
            #print(f"Django Admin/Superuser account created with username: {os.getenv('DJANGO_SUPERUSER', 'admin')} and pw: {os.getenv('DJANGO_SUPERUSER_PASSWORD', 'yxcv4321')}")
            print(f"Django Admin/Superuser account created")
        else:
            print('Admin accounts can only be initialized if no Accounts exist')



"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_data filter app *

:details: lara_django_data filter app.
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""

import django_filters
from django.db.models import Q
from django_filters.rest_framework import FilterSet

from .models import (
    Address,
    City,
    Country,
    CountrySubdivision,
    Currency,
    DataType,
    ExternalResource,
    ExtraData,
    GeoLocation,
    ItemStatus,
    LARASyncServer,
    Location,
    MediaType,
    Namespace,
    PhysicalStateMatter,
    PhysicalUnit,
    Room,
    RoomCategory,
    ScientificConstant,
    SubdivisionType,
    Tag,
)


class MediaTypeFilterSet(FilterSet):
    class Meta:
        model = MediaType
        fields = {
            "name": ["exact", "contains"],
            "media_type": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class DataTypeFilterSet(FilterSet):
    class Meta:
        model = DataType
        fields = {
            "name": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class NamespaceFilterSet(FilterSet):
    # uri =CharFilter(field_name="uri", lookup_expr="contains")
    class Meta:
        model = Namespace
        fields = {
            "uri": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class TagFilterSet(FilterSet):
    class Meta:
        model = Tag
        fields = {
            "tag": ["exact", "contains"],
            "iri": ["exact", "contains"],
        }


class ItemStatusFilterSet(FilterSet):
    class Meta:
        model = ItemStatus
        fields = {
            "status": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class ExtraDataFilterSet(FilterSet):
    class Meta:
        model = ExtraData
        fields = {
            "namespace__uri": ["exact", "contains"],
            "name_display": ["exact", "contains"],
            "name": ["exact", "contains"],
            "name_full": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "url": ["exact", "contains"],
            "datetime_created": ["lt", "gt", "exact", "range"],
            "datetime_last_modified": ["lt", "gt", "exact", "range"],
            "description": ["exact", "contains"],
        }


class ExternalResourceFilterSet(FilterSet):
    """This filter is used to filter ExternalResource objects by name, description and URL."""

    class Meta:
        model = ExternalResource
        fields = {
            "name": ["exact", "contains"],
            "name_display": ["exact", "contains"],
            "url": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "description": ["exact", "contains"],
            "datetime_last_modified": ["lt", "gt", "exact", "range"],
        }


class PhysicalUnitFilterSet(FilterSet):
    class Meta:
        model = PhysicalUnit
        fields = {
            "name": ["exact", "contains"],
            "symbol": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class PhysicalStateMatterFilterSet(FilterSet):
    class Meta:
        model = PhysicalStateMatter
        fields = {
            "name": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class ScientificConstantFilterSet(FilterSet):
    class Meta:
        model = ScientificConstant
        fields = {
            "name": ["exact", "contains"],
            "symbol": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "url_source": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class CurrencyFilterSet(FilterSet):
    class Meta:
        model = Currency
        fields = {
            "name": ["exact", "contains"],
            "symbol": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "datetime_last_modified": ["lt", "gt", "exact", "range"],
        }


class GeoLocationFilterSet(FilterSet):
    class Meta:
        model = GeoLocation
        fields = {
            "name": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "datetime_last_modified": ["lt", "gt", "exact", "range"],
        }


class CountryFilterSet(FilterSet):
    class Meta:
        model = Country
        fields = {
            "name": ["exact", "contains"],
            "name_local": ["exact", "contains"],
            "alpha2code": ["exact", "contains"],
            "alpha3code": ["exact", "contains"],
            "numeric_code": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class SubdivisionTypeFilterSet(FilterSet):
    class Meta:
        model = SubdivisionType
        fields = {
            "name": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class CountrySubdivisionFilterSet(FilterSet):
    class Meta:
        model = CountrySubdivision
        fields = {
            "name": ["exact", "contains"],
            "name_local": ["exact", "contains"],
            "code": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class CityFilterSet(FilterSet):
    class Meta:
        model = City
        fields = {
            "name": ["exact", "contains"],
            "name_local": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class AddressFilterSet(FilterSet):
    name = django_filters.CharFilter(method="filter_name")

    class Meta:
        model = Address
        fields = {
            "namespace__uri": ["exact", "contains"],
            "building": ["exact", "contains"],
            "street": ["exact", "contains"],
            "house_number": ["exact"],
            "postal_code": ["exact"],
            "city__name": ["exact", "contains"],
            "city__name_local": ["exact", "contains"],
            "country__name": ["exact", "contains"],
            "country__name_local": ["exact", "contains"],
            "country__alpha2code": ["exact"],
            "supplement": ["exact", "contains"],
            "description": ["exact", "contains"],
        }
    
    def filter_name(self, queryset, name, value):
        # Simulate searching by the "name" property using the real fields

        tokens = value.strip().split()
        if not tokens:
            return queryset.filter(
                Q(street__icontains=value)
                | Q(building__icontains=value)
                | Q(house_number__icontains=value)
                | Q(postal_code__icontains=value)
                | Q(city__name__icontains=value)
                | Q(country__name__icontains=value)
            )
        for token in tokens:
            queryset = queryset.filter(
                Q(street__icontains=token)
                | Q(building__icontains=token)
                | Q(house_number__icontains=token)
                | Q(postal_code__icontains=token)
                | Q(city__name__icontains=token)
                | Q(country__name__icontains=token)
            )
        return queryset


class RoomCategoryFilterSet(FilterSet):
    class Meta:
        model = RoomCategory
        fields = {
            "category": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class RoomFilterSet(FilterSet):
    class Meta:
        model = Room
        fields = {
            "category__category": ["exact", "contains"],
            "namespace__uri": ["exact", "contains"],
            "name": ["exact", "contains"],
            "number": ["exact", "contains"],
            "floor": ["exact", "contains"],
            "phone_number": ["exact", "contains"],
            "area": ["exact", "contains"],
            "max_people": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class LocationFilterSet(FilterSet):
    class Meta:
        model = Location
        fields = {
            "namespace__uri": ["exact", "contains"],
            "name": ["exact", "contains"],
            "barcode": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class LARASyncServerFilterSet(FilterSet):
    class Meta:
        model = LARASyncServer
        fields = {
            "url": ["exact", "contains"],
            "description": ["exact", "contains"],
        }
